exception Not_implemented

open Form
open FormUtil
open Printf
open FolTranslation
open Common
open MlPrintForm
open Util
open TypeReconstruct



let alloy_print_sequent (sqob : sq_obligation) : bool=
  let _ = Util.msg ("\nLV - BEGIN printing sequent\n")
  in
  let (hyps, goal) = sqob.sqob_sq
  in 
  let _ = Util.msg ("LV[sequent]: " ^ string_of_sequent (hyps, goal) ^ "\n")
  in
  let _ = Util.msg ("\nLV - END printing sequent\n")
  in
    true

let alloy_print_fv (f : form) : string =
  String.concat " || " (fv f)

(** ------------------------------------------------------------------------------------- *)
type vartype = string (** this is only temporary *)

let to_string_ident (i : ident) : string = i

let to_string_simpleType (st : simpleType) : string =
  match st with
    | TypeInt -> "int"
    | _ -> raise Not_implemented

let rec to_string_typeForm (tf : typeForm) : string =
  match tf with 
    | TypeVar v -> Util.replace_dot_with_uscore (to_string_ident v)
(**    | TypeVar v -> (to_string_ident v) *)
    | TypeApp (st, []) -> (to_string_simpleType st)
    | _ -> raise Not_implemented

let to_string_constValue (cv : constValue) : string =
  match cv with
    | Eq -> "=" | Or -> "|" | And -> "&" | Not -> "~" | Impl -> "implies" 
    | Lt -> "<" | Gt -> ">" | LtEq -> "<=" | GtEq -> ">=" | UMinus -> "-"
    | Plus -> "+" | Minus -> "-" | Mult -> "*" | Div -> "div" | Mod -> "mod"
    | IntConst k -> string_of_int k | StringConst s -> "''" ^ s ^ "''"
    | BoolConst true -> "true" | BoolConst false -> "false"
    | StringConst _ -> "" | Comment -> ""
    | Cap -> "&" | Cup -> "+" | Diff -> "-" | Subset -> "in" | Seteq -> "="
    | _ -> "" (* just ignore anything else, for now *)

let to_string_typedIdent (ti : typedIdent) : string =
  let (id, ty) = ti 
  in
    to_string_typeForm ty ^ " " ^ to_string_ident id

let rec to_string_form (tfv_list : (ident * string) list) (f : form) : string =
  match f with
    | Var v -> (" " ^ List.assoc v tfv_list ^ " " ^ (Util.replace_dot_with_uscore v))

(**	(" " ^ List.assoc v tfv_list ^ " " ^ v) *)

    | Const c -> (" " ^ to_string_constValue c)
    | App (f1, fl) -> 
	let rf1 = (" " ^ to_string_form tfv_list f1) 
	in
	let rfl = String.concat rf1 (List.map (to_string_form tfv_list) fl)
	in
	  rfl
    | TypedForm (f1, tf) ->
	let rf1 = (" " ^ to_string_form tfv_list f1)
	in
	let rtf = (" " ^ to_string_typeForm tf)
	in
	  rf1
    | _ -> raise Not_implemented


let do_declaration (p_name : string)
    (tfv_list : (ident * vartype) list) : string =
  let pred_head = "assert " ^ p_name ^ " {"
  in
  let var_decl ((var, typ) : ident * vartype) : string =
    if (String.contains var '.') then 
      String.fill var (String.index var '.') 1 '_';
    "all " ^ var ^ " : " ^ (String.capitalize typ)
  in
  let pred_decl = 
    (String.concat " | " (List.map var_decl tfv_list)) ^ " | "
  in
    pred_head ^ pred_decl
      
let form_to_alloy (pred_name : string) 
                  (typedfv_list : (ident * vartype) list) 
		  (f : form) (scope : int) =
  print_endline (do_declaration pred_name typedfv_list ^ to_string_form typedfv_list f ^ "}");
  print_endline ("check " ^ pred_name ^ " for " ^ string_of_int scope)
 
let form_to_alloy_string (pred_name : string)
                         (typedfv_list : (ident * vartype) list)
			 (f : form) (scope : int) (bitlength : int) : string =
  let r1 = do_declaration pred_name typedfv_list ^ to_string_form typedfv_list f ^ "}"
  in
  let r2 = "check " ^ pred_name ^" for "^ string_of_int scope ^ string_of_int bitlength ^ " int"
  in
  let outs = r1 ^ r2
  in
    outs

(** generate the alloy .als model file 
    alloy_model_name -> output filename
    alloy_model -> string representation of alloy model
    output is unit
*)
let generate_als (alloy_model_name : string) (alloy_model : string) =
  let _ = Util.msg ("\nGenerating Alloy model\n")
  in
  let _ =
    let oc = open_out (alloy_model_name ^ ".als" )
    in    
      Printf.fprintf oc "%s\n" alloy_model;   
      close_out oc
  in
    ()


(** ------------------------------------------------------------------------------------- *)

let unmk_comment f = 
  match f with
    | App(Const Comment,[Const (StringConst _);f1]) -> f1
    | _ -> f


let alloy_print_sequent_fv (sqob : sq_obligation) (scope : int) (bitlength : int) : bool =
  let (oldhyps, oldgoal) = sqob.sqob_sq
  in
  let hyps = List.map unmk_comment oldhyps
  in
  let goal = unmk_comment oldgoal
  in
  let all_fmls = goal :: hyps
  in
  let all_freevars = List.flatten (List.map fv all_fmls)
  in
  let _ = Util.msg("\nHyps vars:\n" ^ String.concat " | " (List.map alloy_print_fv hyps))
  in
  let _ = Util.msg("\nGoal vars: " ^ alloy_print_fv goal)
  in
  let varty_pair = List.flatten (List.map FormUtil.get_annotated_types all_fmls)
  in
  let fvlist = List.map (fun (x, y) -> (x, to_string_typeForm y)) varty_pair
  in
  let po = mk_impl (smk_metaand hyps, goal)
  in
  let model = form_to_alloy_string "alloy_model" fvlist po scope bitlength
  in
  let _ = Util.msg ("\nAlloy model:\n\n" ^ model ^ "\n\nEnd of alloy model.\n")
  in
    (generate_als "alloy_model" model);
(**    (Util.run_command  
	("java -cp alloy4.jar  edu.mit.csail.sdg.alloy4whole.ExampleUsingTheCompiler" ^
	    " alloy_model.als")); *)
    true

