exception Not_implemented

type ident = string

type vartype = ident

type simpleType = | TypeInt

type typeForm = | TypeVar of ident
		| TypeApp of simpleType * typeForm list

type constValue = Eq | Or | And | Not | Impl | Lt | Gt | LtEq | GtEq
		  | UMinus | Plus | Minus | Mult | Div | Mod
		  | IntConst of int | BoolConst of bool | StringConst of string

type typedIdent = ident * typeForm

type form = | Var of ident
	    | Const of constValue
	    | App of form * form list
	    | TypedForm of form * typeForm
		
let to_string_ident (i : ident) : string = i

let to_string_simpleType (st : simpleType) : string =
  match st with
    | TypeInt -> "int"
    | _ -> raise Not_implemented

let rec to_string_typeForm (tf : typeForm) : string =
  match tf with 
    | TypeVar v -> (to_string_ident v)
    | TypeApp (st, []) -> (to_string_simpleType st)
    | _ -> raise Not_implemented

let to_string_constValue (cv : constValue) : string =
  match cv with
    | Eq -> "=" | Or -> "|" | And -> "&" | Not -> "~" | Impl -> "implies" 
    | Lt -> "<" | Gt -> ">" | LtEq -> "<=" | GtEq -> ">=" | UMinus -> "-"
    | Plus -> "+" | Minus -> "-" | Mult -> "*" | Div -> "div" | Mod -> "mod"
    | IntConst k -> string_of_int k | StringConst s -> "''" ^ s ^ "''"
    | BoolConst true -> "true" | BoolConst false -> "false"
    | _ -> "" (* just ignore anything else, for now *)

let to_string_typedIdent (ti : typedIdent) : string =
  let (id, ty) = ti 
  in
    to_string_typeForm ty ^ " " ^ to_string_ident id

let rec to_string_form (tfv_list : (ident * string) list) (f : form) : string =
  match f with
    | Var v -> (" " ^ List.assoc v tfv_list ^ " " ^ v) 
    | Const c -> (" " ^ to_string_constValue c)
    | App (f1, fl) -> 
	let rf1 = (" " ^ to_string_form tfv_list f1) 
	in
	let rfl = String.concat rf1 (List.map (to_string_form tfv_list) fl)
	in
	  rfl
    | TypedForm (f1, tf) ->
	let rf1 = (" " ^ to_string_form tfv_list f1)
	in
	let rtf = (" " ^ to_string_typeForm tf)
	in
	  rf1            (** "," ^ rtf ^ ")" *)

(** *)
let do_declaration (p_name : string)
    (tfv_list : (ident * vartype) list) : string =
  let pred_head = "assert " ^ p_name ^ "{"
  in
  let var_decl ((var, typ) : ident * vartype) : string =
    "all " ^ var ^ " : " ^ (String.capitalize typ)
  in
  let pred_decl = 
    (String.concat " | " (List.map var_decl tfv_list)) ^ " | "
  in
    pred_head ^ pred_decl
      
let form_to_alloy (pred_name : string) 
                  (typedfv_list : (ident * vartype) list) 
		  (f : form) (scope : int) =
  print_endline (do_declaration pred_name typedfv_list ^ to_string_form typedfv_list f ^ "}");
  print_endline ("check " ^ pred_name ^ " for " ^ string_of_int scope)
 
let form_to_alloy_string (pred_name : string)
                         (typedfv_list : (ident * vartype) list)
			 (f : form) (scope : int) : string =
  let r1 = do_declaration pred_name typedfv_list ^ to_string_form typedfv_list f ^ "}"
  in
  let r2 = "check " ^ pred_name ^ " for " ^ string_of_int scope
  in
  let outs = r1 ^ r2
  in
    outs

(** generate the alloy .als model file 
    alloy_model_name -> output filename
    alloy_model -> string representation of alloy model
    output is unit
*)
let generate_als (alloy_model_name : string) (alloy_model : string) =
  let _ =
    let oc = open_out (alloy_model_name ^ ".als")
    in    
      Printf.fprintf oc "%s\n" alloy_model;   
      close_out oc
  in
    ()


(** invoke alloy using previously generated model file
let run_alloy (invoke_alloy : string) 
              (alloy_model_name : string) 
	      (alloy_model : string) : () =
  let _ = generate_als alloy_model_name alloy_model
  in
  (** eventually read the response somehow from alloy *)
  let alloy_in, alloy_out = Unix.open_process invoke_alloy ^ alloy_model_name ^ ".als"
  in
    ()
*)
(** ------------------------------------------------------------------------ *)
(** TRANSLATION CODE ENDS HERE *)

let fvlist = [("amount", "int"); ("account", "int"); ("checkingBalance","int")];;

let fml = (App((Const Gt),[(TypedForm((Var "account"),(TypeApp(TypeInt,[])))); (TypedForm((Var "amount"),(TypeApp(TypeInt,[]))))]));;
			   
let fml1 = (App((Const Impl), 
[(App((Const Gt),[(TypedForm((Var "amount"),(TypeApp(TypeInt,[])))); 
                  (TypedForm((Const (IntConst 0)),(TypeApp(TypeInt,[]))))])) ; 
 (App((Const Eq),[(TypedForm((App((Const Plus),
                              [(TypedForm((Var "checkingBalance"),(TypeApp(TypeInt,[])))); 
                               (TypedForm((Var "amount"),(TypeApp(TypeInt,[]))))])),
                             (TypeApp(TypeInt,[])))); 
                  (TypedForm((App((Const Plus),
                              [(TypedForm((Var "checkingBalance"),(TypeApp(TypeInt,[])))); 
                               (TypedForm((Const (IntConst 1)),(TypeApp(TypeInt,[]))))])),
                               (TypeApp(TypeInt,[]))))]))
]));;

(**
let rec do_body (tfv_list : (ident * vartype) list)
    (formula : form) : string = 
  match formula with
    | Inteq (ls, rs) -> " " ^ do_body tfv_list ls ^ " =" ^ do_body tfv_list rs 
    | Intplus (ls, rs) -> " " ^ do_body tfv_list ls ^ " +" ^ do_body tfv_list rs
    | Intless (ls, rs) -> " " ^ do_body tfv_list ls ^ " <" ^ do_body tfv_list rs
    | Const c -> " " ^ string_of_int c
    | Var v -> " " ^ List.assoc v tfv_list ^ " " ^ v
    | _ -> raise Not_implemented

;;

let _ = to_string_form fvlist fml;;
*)

let _ = form_to_alloy "deposit" fvlist fml1 10;;

let model = form_to_alloy_string "deposit" fvlist fml1 10;;

generate_als "deposit" model;;

