exception Not_implemented

type vartype = string

type ident = string

type form =
    | Var of ident
    | Const of int
    | Intless of form * form
    | Intplus of form * form
    | Inteq of form * form

let do_declaration (p_name : string)
    (tfv_list : (ident * vartype) list) : string =
  let pred_head = "assert " ^ p_name ^ "(){"
  in
  let var_decl ((var, typ) : ident * vartype) : string =
    "all " ^ var ^ " : " ^ typ
  in
  let pred_decl = 
    (String.concat " | " (List.map var_decl tfv_list)) ^ " | "
  in
    pred_head ^ pred_decl
      
let rec do_body (tfv_list : (ident * vartype) list)
    (formula : form) : string = 
  match formula with
    | Inteq (ls, rs) -> " " ^ do_body tfv_list ls ^ " =" ^ do_body tfv_list rs 
    | Intplus (ls, rs) -> " " ^ do_body tfv_list ls ^ " +" ^ do_body tfv_list rs
    | Intless (ls, rs) -> " " ^ do_body tfv_list ls ^ " <" ^ do_body tfv_list rs
    | Const c -> " " ^ string_of_int c
    | Var v -> " " ^ List.assoc v tfv_list ^ " " ^ v
    | _ -> raise Not_implemented

let form_to_alloy (pred_name : string) 
                  (typedfv_list : (ident * vartype) list) 
		  (f : form) (scope : int) =
  print_endline (do_declaration pred_name typedfv_list ^ do_body typedfv_list f ^ "}");
  print_endline ("check " ^ pred_name ^ " for " ^ string_of_int scope)
  
	

let fvlist = [("amount", "int"); ("account", "int")];;
 
let _ = form_to_alloy "deposit" fvlist (Inteq ((Intplus ((Var "amount"),(Const 1))), (Var "account"))) 10;;
